94057c3d3e
PR Test (NPU) / check-changes (push) Has been cancelled
PR Test (NPU) / pr-gate (push) Has been cancelled
PR Test (NPU) / set-image-config (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-1-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (0) (push) Has been cancelled
PR Test (NPU) / stage-b-test-2-npu-a2 (1) (push) Has been cancelled
PR Test (NPU) / stage-b-test-4-npu-a3 (push) Has been cancelled
PR Test (NPU) / stage-b-test-16-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-1-npu-a3 (push) Has been cancelled
PR Test (NPU) / multimodal-gen-test-2-npu-a3 (push) Has been cancelled
PR Test (Arm64) / pr-gate (push) Has been cancelled
PR Test (Arm64) / check-changes (push) Has been cancelled
PR Test (Arm64) / build-test (push) Has been cancelled
PR Test (sgl-router) / gate (push) Has been cancelled
PR Test (sgl-router) / tier-1 — lint (push) Has been cancelled
PR Test (sgl-router) / tier-2 — build + test (push) Has been cancelled
PR Test (sgl-router) / tier-3 — docker (placeholder) (push) Has been cancelled
PR Test (sgl-router) / tier-3 — k8s integration (push) Has been cancelled
PR Test (sgl-router) / tier-3 — e2e (push) Has been cancelled
PR Test (sgl-router) / finish (push) Has been cancelled
PR Test (NPU) / single-node-poc (map[name:qwen3_6_27b_w8a8_1p_in64k_out1k_50ms runner:linux-aarch64-a3-2 test_case:test/registered/ascend/performance/qwen3_6_27b/test_npu_qwen3_6_27b_w8a8_1p_in64k_out1k_50ms.py test_type:perf]) (push) Has been cancelled
PR Test (NPU) / pr-test-npu-finish (push) Has been cancelled
PR Test (Xeon) / pr-gate (push) Has been cancelled
PR Test (Xeon) / check-changes (push) Has been cancelled
PR Test (Xeon) / build-test (, xeon-gnr, base-b-test-cpu) (push) Has been cancelled
PR Test (XPU) / check-changes (push) Has been cancelled
PR Test (XPU) / pr-gate (push) Has been cancelled
PR Test (XPU) / stage-a-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / wait-for-stage-a (push) Has been cancelled
PR Test (XPU) / stage-b-test-1-gpu-xpu (push) Has been cancelled
PR Test (XPU) / finish (push) Has been cancelled
CI Model Inventory / build-inventory (push) Has been cancelled
Lint / lint (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Compilation Check (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Manual Policy (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark - Request Processing (push) Has been cancelled
PR Benchmark (SMG Components) / Benchmark Summary (push) Has been cancelled
PR Test (SMG) / build-wheel (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on windows (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (x86_64 - auto) (push) Has been cancelled
PR Test (SMG) / python-unit-tests (push) Has been cancelled
PR Test (SMG) / unit-tests (push) Has been cancelled
PR Test (SMG) / benchmarks (push) Has been cancelled
PR Test (SMG) / chat-completions (push) Has been cancelled
PR Test (SMG) / chat-completions-4gpu (push) Has been cancelled
PR Test (SMG) / e2e (push) Has been cancelled
PR Test (SMG) / docker-build-test (push) Has been cancelled
PR Test (SMG) / k8s-integration (push) Has been cancelled
PR Test (SMG) / finish (push) Has been cancelled
PR Test (SMG) / summarize-benchmarks (push) Has been cancelled
Release SGLang Model Gateway Docker Image / publish (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on macos (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - auto) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (aarch64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / build on linux (x86_64 - musllinux_1_1) (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Build SDist (push) Has been cancelled
Release SGLang Model Gateway to PyPI / Upload to PyPI (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (aarch64, 12.9, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu129-matrix (x86_64, 12.9, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu129 (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (aarch64, 13.0, 3.10, arm-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / build-cu130-matrix (x86_64, 13.0, 3.10, x64-kernel-build-node) (push) Has been cancelled
Release SGLang Kernels / release-cu130 (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 700) (push) Has been cancelled
Release SGLang Kernels / build-rocm-matrix (3.10, 720) (push) Has been cancelled
Release SGLang Kernels / release-rocm700 (push) Has been cancelled
Release SGLang Kernels / release-rocm720 (push) Has been cancelled
Release SGLang Kernels / build-musa43 (43, 3.10) (push) Has been cancelled
Release SGLang Kernels / release-musa43 (push) Has been cancelled
1205 lines
41 KiB
Python
1205 lines
41 KiB
Python
# Adapt from https://github.com/fla-org/flash-linear-attention/blob/main/fla/ops/gated_delta_rule/fused_recurrent.py
|
|
# -*- coding: utf-8 -*-
|
|
# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang
|
|
|
|
from typing import Optional, Tuple
|
|
|
|
import torch
|
|
import triton
|
|
import triton.language as tl
|
|
|
|
from sglang.srt.layers.attention.fla.op import exp
|
|
from sglang.srt.layers.attention.fla.utils import input_guard
|
|
|
|
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def fused_recurrent_gated_delta_rule_fwd_kernel(
|
|
q,
|
|
k,
|
|
v,
|
|
g,
|
|
beta,
|
|
o,
|
|
h0,
|
|
ht,
|
|
cu_seqlens,
|
|
scale,
|
|
T,
|
|
B: tl.constexpr,
|
|
H: tl.constexpr,
|
|
HV: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
USE_INITIAL_STATE: tl.constexpr, # whether to use initial state
|
|
STORE_FINAL_STATE: tl.constexpr, # whether to store final state
|
|
IS_BETA_HEADWISE: tl.constexpr, # whether beta is headwise vector or scalar,
|
|
USE_QK_L2NORM_IN_KERNEL: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
IS_KDA: tl.constexpr,
|
|
):
|
|
i_k, i_v, i_nh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_n, i_hv = i_nh // HV, i_nh % HV
|
|
i_h = i_hv // (HV // H)
|
|
if IS_VARLEN:
|
|
bos, eos = tl.load(cu_seqlens + i_n).to(tl.int64), tl.load(
|
|
cu_seqlens + i_n + 1
|
|
).to(tl.int64)
|
|
all = T
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_n * T, i_n * T + T
|
|
all = B * T
|
|
o_k = i_k * BK + tl.arange(0, BK)
|
|
o_v = i_v * BV + tl.arange(0, BV)
|
|
|
|
p_q = q + (bos * H + i_h) * K + o_k
|
|
p_k = k + (bos * H + i_h) * K + o_k
|
|
p_v = v + (bos * HV + i_hv) * V + o_v
|
|
if IS_BETA_HEADWISE:
|
|
p_beta = beta + (bos * HV + i_hv) * V + o_v
|
|
else:
|
|
p_beta = beta + bos * HV + i_hv
|
|
if not IS_KDA:
|
|
p_g = g + bos * HV + i_hv
|
|
else:
|
|
p_gk = g + (bos * H + i_h) * K + o_k
|
|
|
|
p_o = o + ((i_k * all + bos) * HV + i_hv) * V + o_v
|
|
|
|
mask_k = o_k < K
|
|
mask_v = o_v < V
|
|
mask_h = mask_v[:, None] & mask_k[None, :]
|
|
|
|
b_h = tl.zeros([BV, BK], dtype=tl.float32)
|
|
if USE_INITIAL_STATE:
|
|
p_h0 = h0 + i_nh * V * K + o_v[:, None] * K + o_k[None, :]
|
|
b_h += tl.load(p_h0, mask=mask_h, other=0).to(tl.float32)
|
|
|
|
for _ in range(0, T):
|
|
b_q = tl.load(p_q, mask=mask_k, other=0).to(tl.float32)
|
|
b_k = tl.load(p_k, mask=mask_k, other=0).to(tl.float32)
|
|
b_v = tl.load(p_v, mask=mask_v, other=0).to(tl.float32)
|
|
|
|
if USE_QK_L2NORM_IN_KERNEL:
|
|
b_q = b_q / (tl.sqrt(tl.sum(b_q * b_q) + 1e-6))
|
|
b_k = b_k / (tl.sqrt(tl.sum(b_k * b_k) + 1e-6))
|
|
b_q = b_q * scale
|
|
# [BV, BK]
|
|
if not IS_KDA:
|
|
b_g = tl.load(p_g).to(tl.float32)
|
|
b_h *= exp(b_g)
|
|
else:
|
|
b_gk = tl.load(p_gk, mask=mask_k, other=0).to(tl.float32)
|
|
b_h *= exp(b_gk[None, :])
|
|
# [BV]
|
|
b_v -= tl.sum(b_h * b_k[None, :], 1)
|
|
if IS_BETA_HEADWISE:
|
|
b_beta = tl.load(p_beta, mask=mask_v, other=0).to(tl.float32)
|
|
else:
|
|
b_beta = tl.load(p_beta).to(tl.float32)
|
|
b_v *= b_beta
|
|
# [BV, BK]
|
|
b_h += b_v[:, None] * b_k[None, :]
|
|
# [BV]
|
|
b_o = tl.sum(b_h * b_q[None, :], 1)
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), mask=mask_v)
|
|
|
|
p_q += H * K
|
|
p_k += H * K
|
|
p_o += HV * V
|
|
p_v += HV * V
|
|
if not IS_KDA:
|
|
p_g += HV
|
|
else:
|
|
p_gk += H * K
|
|
p_beta += HV * (V if IS_BETA_HEADWISE else 1)
|
|
|
|
if STORE_FINAL_STATE:
|
|
p_ht = ht + i_nh * V * K + o_v[:, None] * K + o_k[None, :]
|
|
tl.store(p_ht, b_h.to(p_ht.dtype.element_ty), mask=mask_h)
|
|
|
|
|
|
def fused_recurrent_gated_delta_rule_fwd(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
output_final_state: bool,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
) -> Tuple[torch.Tensor, torch.Tensor]:
|
|
B, T, H, K, V = *k.shape, v.shape[-1]
|
|
HV = v.shape[2]
|
|
N = B if cu_seqlens is None else len(cu_seqlens) - 1
|
|
BK, BV = triton.next_power_of_2(K), min(triton.next_power_of_2(V), 32)
|
|
NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
|
|
assert NK == 1, "NK > 1 is not supported yet"
|
|
num_stages = 3
|
|
num_warps = 1
|
|
|
|
o = q.new_empty(NK, *v.shape)
|
|
if output_final_state:
|
|
final_state = q.new_empty(N, HV, V, K, dtype=torch.float32)
|
|
else:
|
|
final_state = None
|
|
|
|
grid = (NK, NV, N * HV)
|
|
fused_recurrent_gated_delta_rule_fwd_kernel[grid](
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
g=g,
|
|
beta=beta,
|
|
o=o,
|
|
h0=initial_state,
|
|
ht=final_state,
|
|
cu_seqlens=cu_seqlens,
|
|
scale=scale,
|
|
T=T,
|
|
B=B,
|
|
H=H,
|
|
HV=HV,
|
|
K=K,
|
|
V=V,
|
|
BK=BK,
|
|
BV=BV,
|
|
USE_INITIAL_STATE=initial_state is not None,
|
|
STORE_FINAL_STATE=final_state is not None,
|
|
IS_BETA_HEADWISE=beta.ndim == v.ndim,
|
|
USE_QK_L2NORM_IN_KERNEL=use_qk_l2norm_in_kernel,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
IS_KDA=False,
|
|
num_warps=num_warps,
|
|
num_stages=num_stages,
|
|
)
|
|
o = o.squeeze(0)
|
|
return o, final_state
|
|
|
|
|
|
# Adapted from vllm project.
|
|
@triton.jit
|
|
def fused_recurrent_gated_delta_rule_packed_decode_kernel(
|
|
mixed_qkv,
|
|
a,
|
|
b,
|
|
A_log,
|
|
dt_bias,
|
|
o,
|
|
h0,
|
|
ht,
|
|
ssm_state_indices,
|
|
scale,
|
|
stride_mixed_qkv_tok: tl.constexpr,
|
|
stride_a_tok: tl.constexpr,
|
|
stride_b_tok: tl.constexpr,
|
|
stride_init_state_token: tl.constexpr,
|
|
stride_final_state_token: tl.constexpr,
|
|
stride_indices_seq: tl.constexpr,
|
|
H: tl.constexpr,
|
|
HV: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
SOFTPLUS_THRESHOLD: tl.constexpr,
|
|
USE_QK_L2NORM_IN_KERNEL: tl.constexpr,
|
|
):
|
|
i_v, i_nh = tl.program_id(0), tl.program_id(1)
|
|
i_n, i_hv = i_nh // HV, i_nh % HV
|
|
i_h = i_hv // (HV // H)
|
|
|
|
o_k = tl.arange(0, BK)
|
|
o_v = i_v * BV + tl.arange(0, BV)
|
|
mask_k = o_k < K
|
|
mask_v = o_v < V
|
|
mask_h = mask_v[:, None] & mask_k[None, :]
|
|
|
|
state_idx = tl.load(ssm_state_indices + i_n * stride_indices_seq).to(tl.int64)
|
|
p_o = o + (i_n * HV + i_hv) * V + o_v
|
|
|
|
if state_idx < 0:
|
|
zero = tl.zeros([BV], dtype=tl.float32).to(p_o.dtype.element_ty)
|
|
tl.store(p_o, zero, mask=mask_v)
|
|
return
|
|
|
|
p_h0 = h0 + state_idx * stride_init_state_token
|
|
p_h0 = p_h0 + i_hv * V * K + o_v[:, None] * K + o_k[None, :]
|
|
b_h = tl.load(p_h0, mask=mask_h, other=0).to(tl.float32)
|
|
|
|
p_mixed = mixed_qkv + i_n * stride_mixed_qkv_tok
|
|
q_off = i_h * K + o_k
|
|
k_off = (H * K) + i_h * K + o_k
|
|
v_off = (2 * H * K) + i_hv * V + o_v
|
|
b_q = tl.load(p_mixed + q_off, mask=mask_k, other=0).to(tl.float32)
|
|
b_k = tl.load(p_mixed + k_off, mask=mask_k, other=0).to(tl.float32)
|
|
b_v = tl.load(p_mixed + v_off, mask=mask_v, other=0).to(tl.float32)
|
|
|
|
if USE_QK_L2NORM_IN_KERNEL:
|
|
b_q = b_q / tl.sqrt(tl.sum(b_q * b_q) + 1e-6)
|
|
b_k = b_k / tl.sqrt(tl.sum(b_k * b_k) + 1e-6)
|
|
b_q = b_q * scale
|
|
|
|
a_val = tl.load(a + i_n * stride_a_tok + i_hv).to(tl.float32)
|
|
b_val = tl.load(b + i_n * stride_b_tok + i_hv).to(tl.float32)
|
|
A_log_val = tl.load(A_log + i_hv).to(tl.float32)
|
|
dt_bias_val = tl.load(dt_bias + i_hv).to(tl.float32)
|
|
x = a_val + dt_bias_val
|
|
softplus_x = tl.where(x <= SOFTPLUS_THRESHOLD, tl.log(1.0 + tl.exp(x)), x)
|
|
g_val = -tl.exp(A_log_val) * softplus_x
|
|
beta_val = tl.sigmoid(b_val).to(b.dtype.element_ty).to(tl.float32)
|
|
|
|
b_h *= exp(g_val)
|
|
b_v -= tl.sum(b_h * b_k[None, :], 1)
|
|
b_v *= beta_val
|
|
b_h += b_v[:, None] * b_k[None, :]
|
|
b_o = tl.sum(b_h * b_q[None, :], 1)
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), mask=mask_v)
|
|
|
|
p_ht = ht + state_idx * stride_final_state_token
|
|
p_ht = p_ht + i_hv * V * K + o_v[:, None] * K + o_k[None, :]
|
|
tl.store(p_ht, b_h.to(p_ht.dtype.element_ty), mask=mask_h)
|
|
|
|
|
|
def fused_recurrent_gated_delta_rule_packed_decode(
|
|
mixed_qkv: torch.Tensor,
|
|
a: torch.Tensor,
|
|
b: torch.Tensor,
|
|
A_log: torch.Tensor,
|
|
dt_bias: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
out: torch.Tensor,
|
|
ssm_state_indices: torch.Tensor,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
if mixed_qkv.ndim != 2:
|
|
raise ValueError(
|
|
f"`mixed_qkv` must be a 2D tensor (got ndim={mixed_qkv.ndim})."
|
|
)
|
|
if mixed_qkv.stride(-1) != 1:
|
|
raise ValueError("`mixed_qkv` must be contiguous in the last dim.")
|
|
if a.ndim != 2 or b.ndim != 2:
|
|
raise ValueError(
|
|
f"`a` and `b` must be 2D tensors (got a.ndim={a.ndim}, b.ndim={b.ndim})."
|
|
)
|
|
if a.stride(-1) != 1 or b.stride(-1) != 1:
|
|
raise ValueError("`a`/`b` must be contiguous in the last dim.")
|
|
if A_log.ndim != 1 or dt_bias.ndim != 1:
|
|
raise ValueError("`A_log`/`dt_bias` must be 1D tensors.")
|
|
if A_log.stride(0) != 1 or dt_bias.stride(0) != 1:
|
|
raise ValueError("`A_log`/`dt_bias` must be contiguous.")
|
|
if ssm_state_indices.ndim != 1:
|
|
raise ValueError(
|
|
f"`ssm_state_indices` must be 1D for packed decode (got ndim={ssm_state_indices.ndim})."
|
|
)
|
|
if not out.is_contiguous():
|
|
raise ValueError("`out` must be contiguous.")
|
|
|
|
dev = mixed_qkv.device
|
|
if any(
|
|
t.device != dev
|
|
for t in (a, b, A_log, dt_bias, initial_state, out, ssm_state_indices)
|
|
):
|
|
raise ValueError("All inputs must be on the same device.")
|
|
|
|
B = mixed_qkv.shape[0]
|
|
if a.shape[0] != B or b.shape[0] != B:
|
|
raise ValueError(
|
|
"Mismatched batch sizes: "
|
|
f"mixed_qkv.shape[0]={B}, a.shape[0]={a.shape[0]}, b.shape[0]={b.shape[0]}."
|
|
)
|
|
if ssm_state_indices.shape[0] != B:
|
|
raise ValueError(
|
|
f"`ssm_state_indices` must have shape [B] (got {tuple(ssm_state_indices.shape)}; expected ({B},))."
|
|
)
|
|
|
|
if initial_state.ndim != 4:
|
|
raise ValueError(
|
|
f"`initial_state` must be a 4D tensor (got ndim={initial_state.ndim})."
|
|
)
|
|
if initial_state.stride(-1) != 1:
|
|
raise ValueError("`initial_state` must be contiguous in the last dim.")
|
|
HV, V, K = initial_state.shape[-3:]
|
|
if a.shape[1] != HV or b.shape[1] != HV:
|
|
raise ValueError(
|
|
f"`a`/`b` must have shape [B, HV] with HV={HV} (got a.shape={tuple(a.shape)}, b.shape={tuple(b.shape)})."
|
|
)
|
|
if A_log.numel() != HV or dt_bias.numel() != HV:
|
|
raise ValueError(
|
|
f"`A_log` and `dt_bias` must have {HV} elements (got A_log.numel()={A_log.numel()}, dt_bias.numel()={dt_bias.numel()})."
|
|
)
|
|
if out.shape != (B, 1, HV, V):
|
|
raise ValueError(
|
|
f"`out` must have shape {(B, 1, HV, V)} (got out.shape={tuple(out.shape)})."
|
|
)
|
|
|
|
qkv_dim = mixed_qkv.shape[1]
|
|
qk_dim = qkv_dim - HV * V
|
|
if qk_dim <= 0 or qk_dim % 2 != 0:
|
|
raise ValueError(
|
|
f"Invalid packed `mixed_qkv` last dim={qkv_dim} for HV={HV}, V={V}."
|
|
)
|
|
q_dim = qk_dim // 2
|
|
if q_dim % K != 0:
|
|
raise ValueError(f"Invalid packed Q size {q_dim}: must be divisible by K={K}.")
|
|
H = q_dim // K
|
|
if H <= 0 or HV % H != 0:
|
|
raise ValueError(
|
|
f"Invalid head config inferred from mixed_qkv: H={H}, HV={HV}."
|
|
)
|
|
|
|
BK = triton.next_power_of_2(K)
|
|
if triton.cdiv(K, BK) != 1:
|
|
raise ValueError(
|
|
f"Packed decode kernel only supports NK=1 (got K={K}, BK={BK})."
|
|
)
|
|
BV = min(triton.next_power_of_2(V), 32)
|
|
num_stages = 3
|
|
num_warps = 1
|
|
|
|
stride_mixed_qkv_tok = mixed_qkv.stride(0)
|
|
stride_a_tok = a.stride(0)
|
|
stride_b_tok = b.stride(0)
|
|
stride_init_state_token = initial_state.stride(0)
|
|
stride_final_state_token = initial_state.stride(0)
|
|
stride_indices_seq = ssm_state_indices.stride(0)
|
|
|
|
NV = triton.cdiv(V, BV)
|
|
grid = (NV, B * HV)
|
|
fused_recurrent_gated_delta_rule_packed_decode_kernel[grid](
|
|
mixed_qkv=mixed_qkv,
|
|
a=a,
|
|
b=b,
|
|
A_log=A_log,
|
|
dt_bias=dt_bias,
|
|
o=out,
|
|
h0=initial_state,
|
|
ht=initial_state,
|
|
ssm_state_indices=ssm_state_indices,
|
|
scale=scale,
|
|
stride_mixed_qkv_tok=stride_mixed_qkv_tok,
|
|
stride_a_tok=stride_a_tok,
|
|
stride_b_tok=stride_b_tok,
|
|
stride_init_state_token=stride_init_state_token,
|
|
stride_final_state_token=stride_final_state_token,
|
|
stride_indices_seq=stride_indices_seq,
|
|
H=H,
|
|
HV=HV,
|
|
K=K,
|
|
V=V,
|
|
BK=BK,
|
|
BV=BV,
|
|
SOFTPLUS_THRESHOLD=20.0,
|
|
USE_QK_L2NORM_IN_KERNEL=use_qk_l2norm_in_kernel,
|
|
num_warps=num_warps,
|
|
num_stages=num_stages,
|
|
)
|
|
return out, initial_state
|
|
|
|
|
|
@triton.jit
|
|
def fused_recurrent_kda_packed_decode_kernel(
|
|
mixed_qkv,
|
|
a,
|
|
b,
|
|
A_log,
|
|
dt_bias,
|
|
o,
|
|
h0,
|
|
ht,
|
|
ssm_state_indices,
|
|
scale,
|
|
stride_mixed_qkv_tok: tl.constexpr,
|
|
stride_a_tok: tl.constexpr,
|
|
stride_b_tok: tl.constexpr,
|
|
stride_init_state_token: tl.constexpr,
|
|
stride_final_state_token: tl.constexpr,
|
|
stride_indices_seq: tl.constexpr,
|
|
H: tl.constexpr,
|
|
HV: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
SOFTPLUS_THRESHOLD: tl.constexpr,
|
|
USE_QK_L2NORM_IN_KERNEL: tl.constexpr,
|
|
):
|
|
"""KDA packed decode: same shape as the GDN packed decode kernel, but
|
|
with a per-K gate (``a`` is ``[B, HV*K]`` and ``dt_bias`` is ``[HV*K]``),
|
|
so the state decay is a per-K vector ``exp(g)`` rather than a scalar."""
|
|
i_v, i_nh = tl.program_id(0), tl.program_id(1)
|
|
i_n, i_hv = i_nh // HV, i_nh % HV
|
|
i_h = i_hv // (HV // H)
|
|
|
|
o_k = tl.arange(0, BK)
|
|
o_v = i_v * BV + tl.arange(0, BV)
|
|
mask_k = o_k < K
|
|
mask_v = o_v < V
|
|
mask_h = mask_v[:, None] & mask_k[None, :]
|
|
|
|
state_idx = tl.load(ssm_state_indices + i_n * stride_indices_seq).to(tl.int64)
|
|
p_o = o + (i_n * HV + i_hv) * V + o_v
|
|
|
|
if state_idx < 0:
|
|
zero = tl.zeros([BV], dtype=tl.float32).to(p_o.dtype.element_ty)
|
|
tl.store(p_o, zero, mask=mask_v)
|
|
return
|
|
|
|
p_h0 = h0 + state_idx * stride_init_state_token
|
|
p_h0 = p_h0 + i_hv * V * K + o_v[:, None] * K + o_k[None, :]
|
|
b_h = tl.load(p_h0, mask=mask_h, other=0).to(tl.float32)
|
|
|
|
p_mixed = mixed_qkv + i_n * stride_mixed_qkv_tok
|
|
q_off = i_h * K + o_k
|
|
k_off = (H * K) + i_h * K + o_k
|
|
v_off = (2 * H * K) + i_hv * V + o_v
|
|
b_q = tl.load(p_mixed + q_off, mask=mask_k, other=0).to(tl.float32)
|
|
b_k = tl.load(p_mixed + k_off, mask=mask_k, other=0).to(tl.float32)
|
|
b_v = tl.load(p_mixed + v_off, mask=mask_v, other=0).to(tl.float32)
|
|
|
|
if USE_QK_L2NORM_IN_KERNEL:
|
|
b_q = b_q / tl.sqrt(tl.sum(b_q * b_q) + 1e-6)
|
|
b_k = b_k / tl.sqrt(tl.sum(b_k * b_k) + 1e-6)
|
|
b_q = b_q * scale
|
|
|
|
# KDA per-K gate: load BK values of ``a`` and ``dt_bias`` for this head.
|
|
p_a = a + i_n * stride_a_tok + i_hv * K + o_k
|
|
p_dt = dt_bias + i_hv * K + o_k
|
|
b_a = tl.load(p_a, mask=mask_k, other=0).to(tl.float32)
|
|
b_dt = tl.load(p_dt, mask=mask_k, other=0).to(tl.float32)
|
|
A_log_val = tl.load(A_log + i_hv).to(tl.float32)
|
|
|
|
x = b_a + b_dt
|
|
softplus_x = tl.where(x <= SOFTPLUS_THRESHOLD, tl.log(1.0 + tl.exp(x)), x)
|
|
b_g = -tl.exp(A_log_val) * softplus_x # [BK]
|
|
|
|
b_val = tl.load(b + i_n * stride_b_tok + i_hv).to(tl.float32)
|
|
# Keep beta in fp32 (no bf16 round-trip) to match the generic decode
|
|
# kernel `fused_sigmoid_gating_delta_rule_update`, which is the reference
|
|
# validated against torch.
|
|
beta_val = tl.sigmoid(b_val).to(tl.float32)
|
|
|
|
# Per-K decay: each K-row of the [V, K] state decays by its own exp(g_k).
|
|
b_h *= exp(b_g)[None, :]
|
|
b_v -= tl.sum(b_h * b_k[None, :], 1)
|
|
b_v *= beta_val
|
|
b_h += b_v[:, None] * b_k[None, :]
|
|
b_o = tl.sum(b_h * b_q[None, :], 1)
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), mask=mask_v)
|
|
|
|
p_ht = ht + state_idx * stride_final_state_token
|
|
p_ht = p_ht + i_hv * V * K + o_v[:, None] * K + o_k[None, :]
|
|
tl.store(p_ht, b_h.to(p_ht.dtype.element_ty), mask=mask_h)
|
|
|
|
|
|
def fused_recurrent_kda_packed_decode(
|
|
mixed_qkv: torch.Tensor,
|
|
a: torch.Tensor,
|
|
b: torch.Tensor,
|
|
A_log: torch.Tensor,
|
|
dt_bias: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
out: torch.Tensor,
|
|
ssm_state_indices: torch.Tensor,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
) -> tuple[torch.Tensor, torch.Tensor]:
|
|
"""KDA T=1 decode fast path. Mirrors ``fused_recurrent_gated_delta_rule_packed_decode``
|
|
but the gate ``g`` is a per-K vector instead of a scalar.
|
|
|
|
Args:
|
|
mixed_qkv: ``[B, 2*H*K + HV*V]`` packed projection output after conv1d.
|
|
Requires ``num_q_heads == num_k_heads == H`` and ``head_q_dim == head_k_dim == K``.
|
|
a: ``[B, HV*K]`` per-K gate input (typically reshaped from ``[B, HV, K]``).
|
|
b: ``[B, HV]`` beta input (post-sigmoid scalar per head).
|
|
A_log: ``[HV]`` log-space decay parameter.
|
|
dt_bias: ``[HV*K]`` per-K time-step bias.
|
|
scale: attention scale factor (typically ``head_k_dim ** -0.5``).
|
|
initial_state: ``[num_slots, HV, V, K]`` full state pool, updated in place.
|
|
out: ``[B, 1, HV, V]`` contiguous output buffer.
|
|
ssm_state_indices: ``[B]`` per-request state slot indices (-1 = skip).
|
|
use_qk_l2norm_in_kernel: apply per-head L2 norm to Q/K inside the kernel.
|
|
"""
|
|
if mixed_qkv.ndim != 2:
|
|
raise ValueError(
|
|
f"`mixed_qkv` must be a 2D tensor (got ndim={mixed_qkv.ndim})."
|
|
)
|
|
if mixed_qkv.stride(-1) != 1:
|
|
raise ValueError("`mixed_qkv` must be contiguous in the last dim.")
|
|
if a.ndim != 2 or b.ndim != 2:
|
|
raise ValueError(
|
|
f"`a` and `b` must be 2D tensors (got a.ndim={a.ndim}, b.ndim={b.ndim})."
|
|
)
|
|
if a.stride(-1) != 1 or b.stride(-1) != 1:
|
|
raise ValueError("`a`/`b` must be contiguous in the last dim.")
|
|
if A_log.ndim != 1 or dt_bias.ndim != 1:
|
|
raise ValueError("`A_log`/`dt_bias` must be 1D tensors.")
|
|
if A_log.stride(0) != 1 or dt_bias.stride(0) != 1:
|
|
raise ValueError("`A_log`/`dt_bias` must be contiguous.")
|
|
if ssm_state_indices.ndim != 1:
|
|
raise ValueError(
|
|
f"`ssm_state_indices` must be 1D for packed decode (got ndim={ssm_state_indices.ndim})."
|
|
)
|
|
if not out.is_contiguous():
|
|
raise ValueError("`out` must be contiguous.")
|
|
|
|
dev = mixed_qkv.device
|
|
if any(
|
|
t.device != dev
|
|
for t in (a, b, A_log, dt_bias, initial_state, out, ssm_state_indices)
|
|
):
|
|
raise ValueError("All inputs must be on the same device.")
|
|
|
|
B = mixed_qkv.shape[0]
|
|
if a.shape[0] != B or b.shape[0] != B:
|
|
raise ValueError(
|
|
"Mismatched batch sizes: "
|
|
f"mixed_qkv.shape[0]={B}, a.shape[0]={a.shape[0]}, b.shape[0]={b.shape[0]}."
|
|
)
|
|
if ssm_state_indices.shape[0] != B:
|
|
raise ValueError(
|
|
f"`ssm_state_indices` must have shape [B] (got {tuple(ssm_state_indices.shape)}; expected ({B},))."
|
|
)
|
|
|
|
if initial_state.ndim != 4:
|
|
raise ValueError(
|
|
f"`initial_state` must be a 4D tensor (got ndim={initial_state.ndim})."
|
|
)
|
|
if initial_state.stride(-1) != 1:
|
|
raise ValueError("`initial_state` must be contiguous in the last dim.")
|
|
HV, V, K = initial_state.shape[-3:]
|
|
if a.shape[1] != HV * K:
|
|
raise ValueError(
|
|
f"`a` must have shape [B, HV*K] with HV={HV}, K={K} "
|
|
f"(got a.shape={tuple(a.shape)})."
|
|
)
|
|
if b.shape[1] != HV:
|
|
raise ValueError(
|
|
f"`b` must have shape [B, HV] with HV={HV} (got b.shape={tuple(b.shape)})."
|
|
)
|
|
if A_log.numel() != HV:
|
|
raise ValueError(f"`A_log` must have {HV} elements (got {A_log.numel()}).")
|
|
if dt_bias.numel() != HV * K:
|
|
raise ValueError(
|
|
f"`dt_bias` must have {HV * K} elements (got {dt_bias.numel()})."
|
|
)
|
|
if out.shape != (B, 1, HV, V):
|
|
raise ValueError(
|
|
f"`out` must have shape {(B, 1, HV, V)} (got out.shape={tuple(out.shape)})."
|
|
)
|
|
|
|
qkv_dim = mixed_qkv.shape[1]
|
|
qk_dim = qkv_dim - HV * V
|
|
if qk_dim <= 0 or qk_dim % 2 != 0:
|
|
raise ValueError(
|
|
f"Invalid packed `mixed_qkv` last dim={qkv_dim} for HV={HV}, V={V}."
|
|
)
|
|
q_dim = qk_dim // 2
|
|
if q_dim % K != 0:
|
|
raise ValueError(
|
|
f"Invalid packed Q size {q_dim}: must be divisible by K={K}. "
|
|
"KDA packed decode requires num_q_heads == num_k_heads and "
|
|
"head_q_dim == head_k_dim."
|
|
)
|
|
H = q_dim // K
|
|
if H <= 0 or HV % H != 0:
|
|
raise ValueError(
|
|
f"Invalid head config inferred from mixed_qkv: H={H}, HV={HV}."
|
|
)
|
|
|
|
BK = triton.next_power_of_2(K)
|
|
if triton.cdiv(K, BK) != 1:
|
|
raise ValueError(
|
|
f"Packed decode kernel only supports NK=1 (got K={K}, BK={BK})."
|
|
)
|
|
BV = min(triton.next_power_of_2(V), 32)
|
|
num_stages = 3
|
|
num_warps = 1
|
|
|
|
stride_mixed_qkv_tok = mixed_qkv.stride(0)
|
|
stride_a_tok = a.stride(0)
|
|
stride_b_tok = b.stride(0)
|
|
stride_init_state_token = initial_state.stride(0)
|
|
stride_final_state_token = initial_state.stride(0)
|
|
stride_indices_seq = ssm_state_indices.stride(0)
|
|
|
|
NV = triton.cdiv(V, BV)
|
|
grid = (NV, B * HV)
|
|
fused_recurrent_kda_packed_decode_kernel[grid](
|
|
mixed_qkv=mixed_qkv,
|
|
a=a,
|
|
b=b,
|
|
A_log=A_log,
|
|
dt_bias=dt_bias,
|
|
o=out,
|
|
h0=initial_state,
|
|
ht=initial_state,
|
|
ssm_state_indices=ssm_state_indices,
|
|
scale=scale,
|
|
stride_mixed_qkv_tok=stride_mixed_qkv_tok,
|
|
stride_a_tok=stride_a_tok,
|
|
stride_b_tok=stride_b_tok,
|
|
stride_init_state_token=stride_init_state_token,
|
|
stride_final_state_token=stride_final_state_token,
|
|
stride_indices_seq=stride_indices_seq,
|
|
H=H,
|
|
HV=HV,
|
|
K=K,
|
|
V=V,
|
|
BK=BK,
|
|
BV=BV,
|
|
SOFTPLUS_THRESHOLD=20.0,
|
|
USE_QK_L2NORM_IN_KERNEL=use_qk_l2norm_in_kernel,
|
|
num_warps=num_warps,
|
|
num_stages=num_stages,
|
|
)
|
|
return out, initial_state
|
|
|
|
|
|
class FusedRecurrentFunction(torch.autograd.Function):
|
|
|
|
@staticmethod
|
|
@input_guard
|
|
def forward(
|
|
ctx,
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state: torch.Tensor,
|
|
output_final_state: bool,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
):
|
|
o, final_state = fused_recurrent_gated_delta_rule_fwd(
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
g=g,
|
|
beta=beta,
|
|
scale=scale,
|
|
initial_state=initial_state,
|
|
output_final_state=output_final_state,
|
|
use_qk_l2norm_in_kernel=use_qk_l2norm_in_kernel,
|
|
cu_seqlens=cu_seqlens,
|
|
)
|
|
|
|
return o, final_state
|
|
|
|
@staticmethod
|
|
@input_guard
|
|
def backward(ctx, do, dht):
|
|
raise NotImplementedError(
|
|
"Backward pass is not implemented yet and we do not have plans to implement it "
|
|
"because we haven't figured out how to compute dg without materializing the full "
|
|
"hidden states for all time steps."
|
|
)
|
|
|
|
|
|
def fused_recurrent_gated_delta_rule(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor = None,
|
|
scale: float = None,
|
|
initial_state: torch.Tensor = None,
|
|
output_final_state: bool = False,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
) -> Tuple[torch.Tensor, torch.Tensor]:
|
|
r"""
|
|
Args:
|
|
q (torch.Tensor):
|
|
queries of shape `[B, T, H, K]`.
|
|
k (torch.Tensor):
|
|
keys of shape `[B, T, H, K]`.
|
|
v (torch.Tensor):
|
|
values of shape `[B, T, HV, V]`.
|
|
GVA is applied if `HV > H`.
|
|
g (torch.Tensor):
|
|
g (decays) of shape `[B, T, HV]`.
|
|
beta (torch.Tensor):
|
|
betas of shape `[B, T, HV]`.
|
|
scale (Optional[int]):
|
|
Scale factor for the RetNet attention scores.
|
|
If not provided, it will default to `1 / sqrt(K)`. Default: `None`.
|
|
initial_state (Optional[torch.Tensor]):
|
|
Initial state of shape `[N, HV, V, K]` for `N` input sequences.
|
|
For equal-length input sequences, `N` equals the batch size `B`.
|
|
Default: `None`.
|
|
output_final_state (Optional[bool]):
|
|
Whether to output the final state of shape `[N, HV, V, K]`. Default: `False`.
|
|
cu_seqlens (torch.LongTensor):
|
|
Cumulative sequence lengths of shape `[N+1]` used for variable-length training,
|
|
consistent with the FlashAttention API.
|
|
Returns:
|
|
o (torch.Tensor):
|
|
Outputs of shape `[B, T, HV, V]`.
|
|
final_state (torch.Tensor):
|
|
Final state of shape `[N, HV, V, K]` if `output_final_state=True` else `None`.
|
|
Examples::
|
|
>>> import torch
|
|
>>> import torch.nn.functional as F
|
|
>>> from einops import rearrange
|
|
>>> from fla.ops.gated_delta_rule import fused_recurrent_gated_delta_rule
|
|
# inputs with equal lengths
|
|
>>> B, T, H, HV, K, V = 4, 2048, 4, 8, 512, 512
|
|
>>> q = torch.randn(B, T, H, K, device='cuda')
|
|
>>> k = F.normalize(torch.randn(B, T, H, K, device='cuda'), p=2, dim=-1)
|
|
>>> v = torch.randn(B, T, HV, V, device='cuda')
|
|
>>> g = F.logsigmoid(torch.rand(B, T, HV, device='cuda'))
|
|
>>> beta = torch.rand(B, T, HV, device='cuda').sigmoid()
|
|
>>> h0 = torch.randn(B, HV, V, K, device='cuda')
|
|
>>> o, ht = fused_gated_recurrent_delta_rule(
|
|
q, k, v, g, beta,
|
|
initial_state=h0,
|
|
output_final_state=True
|
|
)
|
|
# for variable-length inputs, the batch size `B` is expected to be 1 and `cu_seqlens` is required
|
|
>>> q, k, v, g, beta = map(lambda x: rearrange(x, 'b t ... -> 1 (b t) ...'), (q, k, v, g, beta))
|
|
# for a batch with 4 sequences, `cu_seqlens` with 5 start/end positions are expected
|
|
>>> cu_seqlens = q.new_tensor([0, 2048, 4096, 6144, 8192], dtype=torch.long)
|
|
>>> o_var, ht_var = fused_gated_recurrent_delta_rule(
|
|
q, k, v, g, beta,
|
|
initial_state=h0,
|
|
output_final_state=True,
|
|
cu_seqlens=cu_seqlens
|
|
)
|
|
"""
|
|
if cu_seqlens is not None:
|
|
if q.shape[0] != 1:
|
|
raise ValueError(
|
|
f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
|
|
f"Please flatten variable-length inputs before processing."
|
|
)
|
|
if initial_state is not None and initial_state.shape[0] != len(cu_seqlens) - 1:
|
|
raise ValueError(
|
|
f"The number of initial states is expected to be equal to the number of input sequences, "
|
|
f"i.e., {len(cu_seqlens) - 1} rather than {initial_state.shape[0]}."
|
|
)
|
|
if scale is None:
|
|
scale = k.shape[-1] ** -0.5
|
|
else:
|
|
assert scale > 0, "scale must be positive"
|
|
if beta is None:
|
|
beta = torch.ones_like(q[..., 0])
|
|
o, final_state = FusedRecurrentFunction.apply(
|
|
q,
|
|
k,
|
|
v,
|
|
g,
|
|
beta,
|
|
scale,
|
|
initial_state,
|
|
output_final_state,
|
|
cu_seqlens,
|
|
use_qk_l2norm_in_kernel,
|
|
)
|
|
return o, final_state
|
|
|
|
|
|
# HAS_EAGLE_TREE_CUSTOM_ATTN_MASK is added to support eagle tree attention mask
|
|
# retrieve_parent_token_ptr: [N, NP2_T], retrieve_next_sibling_ptr: [N, NP2_T]
|
|
# e.g. for a sequence of length 4, the eagle tree attention structure is:
|
|
# retrieve_next_token=[1, 3, -1, -1] -> retrieve_next_token[i]: the 1st child token of token i
|
|
# retrieve_next_sibling=[-1, 2, -1, -1] -> retrieve_next_sibling[i]: the 1st tree sibling token of token i
|
|
# retrieve_parent_token=[n/a, 0, 0, 1] -> retrieve_parent_token[i]: the parent token of token i
|
|
# Tree:
|
|
# 0
|
|
# / \
|
|
# 1 2
|
|
# /
|
|
# 3
|
|
# When calculating token 3's attention, it should attend to token 1 (parent) and token 0 (grand-parent)
|
|
# When calculating token 2's attention, it should attend to token 0 (parent)
|
|
@triton.jit(do_not_specialize=["T"])
|
|
def fused_recurrent_gated_delta_rule_update_fwd_kernel(
|
|
q,
|
|
k,
|
|
v,
|
|
g,
|
|
beta,
|
|
o,
|
|
h0_source,
|
|
h0_indices,
|
|
cu_seqlens,
|
|
scale,
|
|
intermediate_states_buffer,
|
|
intermediate_state_indices,
|
|
cache_steps,
|
|
retrieve_parent_token_ptr,
|
|
stride_retrieve_parent_token_seq: tl.constexpr,
|
|
stride_retrieve_parent_token_token: tl.constexpr,
|
|
T,
|
|
NP2_T: tl.constexpr,
|
|
B: tl.constexpr,
|
|
H: tl.constexpr,
|
|
HV: tl.constexpr,
|
|
K: tl.constexpr,
|
|
V: tl.constexpr,
|
|
BK: tl.constexpr,
|
|
BV: tl.constexpr,
|
|
USE_INITIAL_STATE: tl.constexpr, # whether to use initial state
|
|
IS_BETA_HEADWISE: tl.constexpr, # whether beta is headwise vector or scalar,
|
|
USE_QK_L2NORM_IN_KERNEL: tl.constexpr,
|
|
IS_VARLEN: tl.constexpr,
|
|
DISABLE_STATE_UPDATE: tl.constexpr, # whether to disable final state update
|
|
DISABLE_OUTPUT_CALCULATION: tl.constexpr, # whether to disable output calculation
|
|
CACHE_INTERMEDIATE_STATES: tl.constexpr,
|
|
HAS_EAGLE_TREE_CUSTOM_ATTN_MASK: tl.constexpr,
|
|
):
|
|
i_k, i_v, i_nh = tl.program_id(0), tl.program_id(1), tl.program_id(2)
|
|
i_n, i_hv = i_nh // HV, i_nh % HV
|
|
i_h = i_hv // (HV // H)
|
|
if IS_VARLEN:
|
|
bos, eos = tl.load(cu_seqlens + i_n).to(tl.int64), tl.load(
|
|
cu_seqlens + i_n + 1
|
|
).to(tl.int64)
|
|
all = T
|
|
T = eos - bos
|
|
else:
|
|
bos, eos = i_n * T, i_n * T + T
|
|
all = B * T
|
|
o_k = i_k * BK + tl.arange(0, BK)
|
|
o_v = i_v * BV + tl.arange(0, BV)
|
|
|
|
p_q = q + (bos * H + i_h) * K + o_k
|
|
p_k = k + (bos * H + i_h) * K + o_k
|
|
p_v = v + (bos * HV + i_hv) * V + o_v
|
|
if IS_BETA_HEADWISE:
|
|
p_beta = beta + (bos * HV + i_hv) * V + o_v
|
|
else:
|
|
p_beta = beta + bos * HV + i_hv
|
|
p_g = g + bos * HV + i_hv
|
|
p_o = o + ((i_k * all + bos) * HV + i_hv) * V + o_v
|
|
|
|
if HAS_EAGLE_TREE_CUSTOM_ATTN_MASK:
|
|
token_indices = tl.arange(0, NP2_T)
|
|
mask_retrieve = token_indices < T
|
|
retrieve_parent_token_base = (
|
|
retrieve_parent_token_ptr
|
|
+ (i_n * stride_retrieve_parent_token_seq)
|
|
+ token_indices * stride_retrieve_parent_token_token
|
|
)
|
|
parent_idx_tokens = tl.load(retrieve_parent_token_base, mask_retrieve)
|
|
|
|
mask_k = o_k < K
|
|
mask_v = o_v < V
|
|
mask_h = mask_v[:, None] & mask_k[None, :]
|
|
|
|
b_h = tl.zeros([BV, BK], dtype=tl.float32)
|
|
if USE_INITIAL_STATE:
|
|
idx = tl.load(h0_indices + i_n)
|
|
# Add bounds checking for idx
|
|
if idx >= 0: # Assuming negative indices are invalid
|
|
p_h0 = (
|
|
h0_source
|
|
+ idx * HV * K * V
|
|
+ i_hv * K * V
|
|
+ o_v[:, None] * K
|
|
+ o_k[None, :]
|
|
)
|
|
b_h += tl.load(p_h0, mask=mask_h, other=0).to(tl.float32)
|
|
|
|
# Prepare intermediate state cache variables if enabled
|
|
cache_idx = -1
|
|
if CACHE_INTERMEDIATE_STATES:
|
|
cache_idx = tl.load(intermediate_state_indices + i_n)
|
|
|
|
step_idx = 0
|
|
for _ in range(0, T):
|
|
if HAS_EAGLE_TREE_CUSTOM_ATTN_MASK:
|
|
# step_idx = 0 should use the b_h from USE_INITIAL_STATE
|
|
if step_idx != 0 and cache_idx >= 0:
|
|
# when calculating current step's attention, load the state from the parent token
|
|
parent_step_idx = tl.sum(
|
|
tl.where(token_indices == step_idx, parent_idx_tokens, 0)
|
|
)
|
|
step_offset = parent_step_idx * HV * K * V
|
|
cache_ptr = (
|
|
intermediate_states_buffer
|
|
+ cache_idx * cache_steps * HV * K * V
|
|
+ step_offset
|
|
+ i_hv * K * V
|
|
+ o_v[:, None] * K
|
|
+ o_k[None, :]
|
|
)
|
|
b_h = tl.load(cache_ptr, mask=mask_h, other=0).to(tl.float32)
|
|
|
|
b_q = tl.load(p_q, mask=mask_k, other=0).to(tl.float32)
|
|
b_k = tl.load(p_k, mask=mask_k, other=0).to(tl.float32)
|
|
b_v = tl.load(p_v, mask=mask_v, other=0).to(tl.float32)
|
|
b_g = tl.load(p_g).to(tl.float32)
|
|
|
|
if USE_QK_L2NORM_IN_KERNEL:
|
|
b_q = b_q / (tl.sqrt(tl.sum(b_q * b_q) + 1e-6))
|
|
b_k = b_k / (tl.sqrt(tl.sum(b_k * b_k) + 1e-6))
|
|
b_q = b_q * scale
|
|
# [BK, BV]
|
|
b_h *= exp(b_g)
|
|
# [BV]
|
|
b_v -= tl.sum(b_h * b_k[None, :], 1)
|
|
if IS_BETA_HEADWISE:
|
|
b_beta = tl.load(p_beta, mask=mask_v, other=0).to(tl.float32)
|
|
else:
|
|
b_beta = tl.load(p_beta).to(tl.float32)
|
|
b_v *= b_beta
|
|
# [BV, BK]
|
|
b_h += b_v[:, None] * b_k[None, :]
|
|
# [BV]
|
|
if not DISABLE_OUTPUT_CALCULATION:
|
|
b_o = tl.sum(b_h * b_q[None, :], 1)
|
|
# core attn output
|
|
tl.store(p_o, b_o.to(p_o.dtype.element_ty), mask=mask_v)
|
|
|
|
# store intermediate states if enabled
|
|
if CACHE_INTERMEDIATE_STATES:
|
|
if cache_idx >= 0:
|
|
# Compute cache pointer for this step
|
|
step_offset = step_idx * HV * K * V
|
|
cache_ptr = (
|
|
intermediate_states_buffer
|
|
+ cache_idx * cache_steps * HV * K * V
|
|
+ step_offset
|
|
+ i_hv * K * V
|
|
+ o_v[:, None] * K
|
|
+ o_k[None, :]
|
|
)
|
|
tl.store(cache_ptr, b_h.to(cache_ptr.dtype.element_ty), mask=mask_h)
|
|
|
|
step_idx += 1
|
|
|
|
p_q += H * K
|
|
p_k += H * K
|
|
p_o += HV * V
|
|
p_v += HV * V
|
|
p_g += HV
|
|
p_beta += HV * (V if IS_BETA_HEADWISE else 1)
|
|
|
|
# Store final state back to h0_source with bounds checking
|
|
# ssm states
|
|
if not DISABLE_STATE_UPDATE:
|
|
idx = tl.load(h0_indices + i_n)
|
|
if idx >= 0: # Add bounds checking
|
|
p_h0 = (
|
|
h0_source
|
|
+ idx * HV * K * V
|
|
+ i_hv * K * V
|
|
+ o_v[:, None] * K
|
|
+ o_k[None, :]
|
|
)
|
|
tl.store(p_h0, b_h.to(p_h0.dtype.element_ty), mask=mask_h)
|
|
|
|
|
|
def fused_recurrent_gated_delta_rule_update_fwd(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state_source: torch.Tensor,
|
|
initial_state_indices: torch.Tensor,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
disable_state_update: bool = False,
|
|
disable_output_calculation: bool = False,
|
|
intermediate_states_buffer: Optional[torch.Tensor] = None,
|
|
intermediate_state_indices: Optional[torch.Tensor] = None,
|
|
cache_steps: Optional[int] = None,
|
|
retrieve_parent_token: Optional[torch.Tensor] = None,
|
|
) -> torch.Tensor:
|
|
B, T, H, K, V = *k.shape, v.shape[-1]
|
|
HV = v.shape[2]
|
|
N = B if cu_seqlens is None else len(cu_seqlens) - 1
|
|
BK, BV = triton.next_power_of_2(K), min(triton.next_power_of_2(V), 8)
|
|
NK, NV = triton.cdiv(K, BK), triton.cdiv(V, BV)
|
|
assert NK == 1, "NK > 1 is not supported yet"
|
|
num_stages = 3
|
|
num_warps = 1
|
|
|
|
if disable_output_calculation:
|
|
# When output calculation is disabled, allocate minimal tensor
|
|
o = q.new_empty(NK, 1, 1, 1, 1) # minimal allocation
|
|
else:
|
|
o = q.new_empty(NK, *v.shape)
|
|
|
|
grid = (NK, NV, N * HV)
|
|
|
|
# prepare retrieve next token buffer strides if provided
|
|
if retrieve_parent_token is not None:
|
|
stride_retrieve_parent_token_seq, stride_retrieve_parent_token_token = (
|
|
retrieve_parent_token.stride(0),
|
|
retrieve_parent_token.stride(1),
|
|
)
|
|
else:
|
|
stride_retrieve_parent_token_seq = stride_retrieve_parent_token_token = 0
|
|
|
|
NP2_T = triton.next_power_of_2(T)
|
|
fused_recurrent_gated_delta_rule_update_fwd_kernel[grid](
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
g=g,
|
|
beta=beta,
|
|
o=o,
|
|
h0_source=initial_state_source,
|
|
h0_indices=initial_state_indices,
|
|
cu_seqlens=cu_seqlens,
|
|
scale=scale,
|
|
intermediate_states_buffer=intermediate_states_buffer,
|
|
intermediate_state_indices=intermediate_state_indices,
|
|
cache_steps=0 if cache_steps is None else cache_steps,
|
|
retrieve_parent_token_ptr=retrieve_parent_token,
|
|
stride_retrieve_parent_token_seq=stride_retrieve_parent_token_seq,
|
|
stride_retrieve_parent_token_token=stride_retrieve_parent_token_token,
|
|
T=T,
|
|
NP2_T=NP2_T,
|
|
B=B,
|
|
H=H,
|
|
HV=HV,
|
|
K=K,
|
|
V=V,
|
|
BK=BK,
|
|
BV=BV,
|
|
USE_INITIAL_STATE=initial_state_source is not None,
|
|
IS_VARLEN=cu_seqlens is not None,
|
|
CACHE_INTERMEDIATE_STATES=intermediate_states_buffer is not None,
|
|
HAS_EAGLE_TREE_CUSTOM_ATTN_MASK=retrieve_parent_token is not None,
|
|
IS_BETA_HEADWISE=beta.ndim == v.ndim,
|
|
USE_QK_L2NORM_IN_KERNEL=use_qk_l2norm_in_kernel,
|
|
DISABLE_STATE_UPDATE=disable_state_update,
|
|
DISABLE_OUTPUT_CALCULATION=disable_output_calculation,
|
|
num_warps=num_warps,
|
|
num_stages=num_stages,
|
|
)
|
|
o = o.squeeze(0)
|
|
return o
|
|
|
|
|
|
class FusedRecurrentUpdateFunction(torch.autograd.Function):
|
|
|
|
@staticmethod
|
|
@input_guard
|
|
def forward(
|
|
ctx,
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor,
|
|
scale: float,
|
|
initial_state_source: torch.Tensor,
|
|
initial_state_indices: torch.Tensor,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
disable_state_update: bool = False,
|
|
disable_output_calculation: bool = False,
|
|
intermediate_states_buffer: Optional[torch.Tensor] = None,
|
|
intermediate_state_indices: Optional[torch.Tensor] = None,
|
|
cache_steps: Optional[int] = None,
|
|
retrieve_parent_token: Optional[torch.Tensor] = None,
|
|
):
|
|
o = fused_recurrent_gated_delta_rule_update_fwd(
|
|
q=q,
|
|
k=k,
|
|
v=v,
|
|
g=g,
|
|
beta=beta,
|
|
scale=scale,
|
|
initial_state_source=initial_state_source,
|
|
initial_state_indices=initial_state_indices,
|
|
use_qk_l2norm_in_kernel=use_qk_l2norm_in_kernel,
|
|
cu_seqlens=cu_seqlens,
|
|
disable_state_update=disable_state_update,
|
|
disable_output_calculation=disable_output_calculation,
|
|
intermediate_states_buffer=intermediate_states_buffer,
|
|
intermediate_state_indices=intermediate_state_indices,
|
|
cache_steps=cache_steps,
|
|
retrieve_parent_token=retrieve_parent_token,
|
|
)
|
|
|
|
return o
|
|
|
|
@staticmethod
|
|
@input_guard
|
|
def backward(ctx, do, dht):
|
|
raise NotImplementedError(
|
|
"Backward pass is not implemented yet and we do not have plans to implement it "
|
|
"because we haven't figured out how to compute dg without materializing the full "
|
|
"hidden states for all time steps."
|
|
)
|
|
|
|
|
|
def fused_recurrent_gated_delta_rule_update(
|
|
q: torch.Tensor,
|
|
k: torch.Tensor,
|
|
v: torch.Tensor,
|
|
g: torch.Tensor,
|
|
beta: torch.Tensor = None,
|
|
scale: float = None,
|
|
initial_state_source: torch.Tensor = None,
|
|
initial_state_indices: torch.Tensor = None,
|
|
cu_seqlens: Optional[torch.LongTensor] = None,
|
|
use_qk_l2norm_in_kernel: bool = False,
|
|
disable_state_update: bool = False,
|
|
disable_output_calculation: bool = False,
|
|
intermediate_states_buffer: Optional[torch.Tensor] = None,
|
|
intermediate_state_indices: Optional[torch.Tensor] = None,
|
|
cache_steps: Optional[int] = None,
|
|
retrieve_parent_token: Optional[torch.Tensor] = None,
|
|
) -> torch.Tensor:
|
|
if cu_seqlens is not None:
|
|
if q.shape[0] != 1:
|
|
raise ValueError(
|
|
f"The batch size is expected to be 1 rather than {q.shape[0]} when using `cu_seqlens`."
|
|
f"Please flatten variable-length inputs before processing."
|
|
)
|
|
if initial_state_source is not None:
|
|
if initial_state_indices.shape[0] != len(cu_seqlens) - 1:
|
|
raise ValueError(
|
|
f"The number of initial states is expected to be equal to the number of input sequences, "
|
|
f"i.e., {len(cu_seqlens) - 1} rather than {initial_state_indices.shape[0]}."
|
|
)
|
|
if initial_state_indices.shape[0] != intermediate_state_indices.shape[0]:
|
|
raise ValueError(
|
|
f"The number of intermediate state indices is expected to be equal to the number of input sequences, "
|
|
f"i.e., {initial_state_indices.shape[0]} != {intermediate_state_indices.shape[0]}."
|
|
)
|
|
if scale is None:
|
|
scale = k.shape[-1] ** -0.5
|
|
else:
|
|
assert scale > 0, "scale must be positive"
|
|
if beta is None:
|
|
beta = torch.ones_like(q[..., 0])
|
|
o = FusedRecurrentUpdateFunction.apply(
|
|
q,
|
|
k,
|
|
v,
|
|
g,
|
|
beta,
|
|
scale,
|
|
initial_state_source,
|
|
initial_state_indices,
|
|
cu_seqlens,
|
|
use_qk_l2norm_in_kernel,
|
|
disable_state_update,
|
|
disable_output_calculation,
|
|
intermediate_states_buffer,
|
|
intermediate_state_indices,
|
|
cache_steps,
|
|
retrieve_parent_token,
|
|
)
|
|
return o
|
|
|
|
|
|
fused_recurrent_gdn = fused_recurrent_gated_delta_rule
|